@@ -1,10 +1,27 @@
 import typing
-from threading import Lock, Event
-from multiprocessing import Value
+from threading import Lock, Event, RLock
+from multiprocessing import Value, Lock as MpLock
 from abc import abstractmethod, ABC
 from .ring_factory import RingFactory, SimpleFactory
 from .exceptions import SequenceNotFound, Empty, SequenceOverwritten, ReadCursorBlock
 
+T = typing.TypeVar("T", bound=typing.Callable[..., typing.Any])
+
+
+def run_with_lock(func: T) -> T:
+    def with_lock(
+        ring_buffer: "RingBufferInternal", *args: typing.Any, **kwargs: typing.Any
+    ) -> T:
+        if ring_buffer._lock is None:
+            return func(ring_buffer, *args, **kwargs)
+        with ring_buffer._lock:
+            return func(ring_buffer, *args, **kwargs)
+
+    return typing.cast(T, with_lock)
+
+
+LockLike = typing.Union[Lock, RLock]
+
 
 class RingBufferInternal:
     def __init__(
@@ -12,6 +29,7 @@ class RingBufferInternal:
         size: int = 16,
         factory: typing.Type[RingFactory] = SimpleFactory,
         cursor_position_value: typing.Union[Value, int] = 0,
+        lock: typing.Optional[LockLike] = None,
     ):
         if not size % 2 == 0:
             raise AttributeError("size must be a factor of 2 for efficient arithmetic.")
@@ -21,6 +39,7 @@ class RingBufferInternal:
 
         self.__cursor_position = cursor_position_value  # position of next write
         self.__ring: typing.List[RingFactory] = [factory() for _ in range(size)]
+        self._lock = lock
 
     def _get_cursor_position(self):
         if isinstance(self.__cursor_position, int):
@@ -34,6 +53,7 @@ class RingBufferInternal:
         else:
             self.__cursor_position.value = value
 
+    @run_with_lock
     def _put(self, value) -> int:
         cursor_position = self._get_cursor_position()
         ring_index = cursor_position % self.ring_size
@@ -44,6 +64,7 @@ class RingBufferInternal:
 
         return cursor_position
 
+    @run_with_lock
     def _get(self, idx: int) -> typing.Tuple[int, typing.Any]:
         cursor_position = self._get_cursor_position()
         if idx >= cursor_position:
@@ -54,6 +75,7 @@ class RingBufferInternal:
 
         return (idx, self.__ring[idx % self.ring_size].get())
 
+    @run_with_lock
     def _get_latest(self) -> typing.Tuple[int, typing.Any]:
         cursor_position = self._get_cursor_position()
         if cursor_position <= 0:
@@ -63,6 +85,7 @@ class RingBufferInternal:
 
         return self._get(idx)
 
+    @run_with_lock
     def _flush(self) -> None:
         self.__ring = [self.factory() for _ in range(self.ring_size)]
         self._set_cursor_position(0)
@@ -106,9 +129,13 @@ class RingBuffer(RingBufferInternal, RandomAccessRingBufferMethods):
         size: int = 16,
         factory: typing.Type[RingFactory] = SimpleFactory,
         cursor_position_value: typing.Union[Value, int] = 0,
+        lock: typing.Optional[LockLike] = None,
     ):
         super().__init__(
-            size=size, factory=factory, cursor_position_value=cursor_position_value
+            size=size,
+            factory=factory,
+            cursor_position_value=cursor_position_value,
+            lock=lock,
         )
 
     def put(self, value):
@@ -129,30 +156,17 @@ class LockedRingBuffer(RingBuffer, RandomAccessRingBufferMethods):
         self,
         size: int = 16,
         factory: typing.Type[RingFactory] = SimpleFactory,
-        lock: Lock = Lock(),
+        lock: RLock = RLock(),
         cursor_position_value: typing.Union[Value, int] = 0,
     ):
         super().__init__(
-            size=size, factory=factory, cursor_position_value=cursor_position_value
+            size=size,
+            factory=factory,
+            cursor_position_value=cursor_position_value,
+            lock=lock,
         )
         self.__lock = lock
 
-    def put(self, value) -> int:
-        with self.__lock:
-            return super().put(value)
-
-    def get(self, idx: int) -> typing.Tuple[int, typing.Any]:
-        with self.__lock:
-            return super().get(idx)
-
-    def get_latest(self) -> typing.Tuple[int, typing.Any]:
-        with self.__lock:
-            return super().get_latest()
-
-    def flush(self) -> None:
-        with self.__lock:
-            return super().flush()
-
 
 class BlockingRingBuffer(RingBufferInternal, SequencedRingBufferMethods):
     _read_cursor = 0
@@ -162,9 +176,13 @@ class BlockingRingBuffer(RingBufferInternal, SequencedRingBufferMethods):
         size: int = 16,
         factory: typing.Type[RingFactory] = SimpleFactory,
         cursor_position_value: typing.Union[Value, int] = 0,
+        lock: typing.Optional[LockLike] = None,
     ):
         super().__init__(
-            size=size, factory=factory, cursor_position_value=cursor_position_value
+            size=size,
+            factory=factory,
+            cursor_position_value=cursor_position_value,
+            lock=lock,
         )
 
     def put(self, value):
@@ -187,26 +205,17 @@ class BlockingLockedRingBuffer(BlockingRingBuffer, SequencedRingBufferMethods):
         self,
         size: int = 16,
         factory: typing.Type[RingFactory] = SimpleFactory,
-        lock: Lock = Lock(),
+        lock: LockLike = RLock(),
         cursor_position_value: typing.Union[Value, int] = 0,
     ):
         super().__init__(
-            size=size, factory=factory, cursor_position_value=cursor_position_value
+            size=size,
+            factory=factory,
+            cursor_position_value=cursor_position_value,
+            lock=lock,
         )
         self.__lock = lock
 
-    def put(self, value):
-        with self.__lock:
-            return super().put(value)
-
-    def next(self):
-        with self.__lock:
-            return super().next()
-
-    def flush(self):
-        with self.__lock:
-            return super().flush()
-
 
 class WaitingBlockingRingBuffer(RingBufferInternal, SequencedRingBufferMethods):
     _read_cursor = 0
@@ -223,7 +232,7 @@ class WaitingBlockingRingBuffer(RingBufferInternal, SequencedRingBufferMethods):
             size=size, factory=factory, cursor_position_value=cursor_position_value
         )
 
-    def put(self, value, timeout: int = None):
+    def put(self, value, timeout: float = None):
         if not self._read_cursor_barrier.is_set():
             self._read_cursor_barrier.set()
         if (self._get_cursor_position() - self._read_cursor) == self.ring_size:
@@ -233,7 +242,7 @@ class WaitingBlockingRingBuffer(RingBufferInternal, SequencedRingBufferMethods):
                 raise ReadCursorBlock()
         return super()._put(value)
 
-    def next(self, timeout: int = None):
+    def next(self, timeout: float = None):
         try:
             res = super()._get(self._read_cursor)
         except SequenceNotFound:
